package za.ac.sun.cs.solver;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.apfloat.Apint;

import za.ac.sun.cs.solver.canonization.Canonizer;
import za.ac.sun.cs.solver.expr.Expression;
import za.ac.sun.cs.solver.expr.IntConstant;
import za.ac.sun.cs.solver.expr.IntVariable;
import za.ac.sun.cs.solver.expr.Operation;
import za.ac.sun.cs.solver.expr.RealVariable;
import za.ac.sun.cs.solver.expr.Variable;
import za.ac.sun.cs.solver.expr.Visitor;
import za.ac.sun.cs.solver.slicing.Slicer;
import za.ac.sun.cs.solver.store.Store;

/**
 * An instance of a solver problem that corresponds to a path condition. It
 * comprises two parts: a reference to a parent instance that contains the
 * "non-fresh" part of the path condition, and an expression that contains the
 * "fresh" part of the path condition which is usually one conjunct that has
 * been added to the path condition.
 * 
 * @author Jaco Geldenhuys <jaco@cs.sun.ac.za>
 */
public class Instance {

	/**
	 * The solver to which this instance belongs. This is used to access the
	 * slicer, canonizer, decision procedure, constraint solver, model counter,
	 * and other instance modifiers.
	 */
	private Solver solver = null;

	/**
	 * The parent instance that contains the non-fresh part of the path
	 * condition.
	 */
	private Instance parent = null;

	/**
	 * The fresh part of the path condition.
	 */
	private Expression expression = null;

	/**
	 * The full expression with the fresh and non-fresh parts of the path
	 * condition.
	 */
	private Expression fullExpression = null;

	/**
	 * The sliced expression. If there is no active slicer, the full expression
	 * is used.
	 */
	private Expression slicedExpression = null;

	/**
	 * The canonized expression. If there is no active canonizer, the sliced
	 * expression is used.
	 */
	private Expression canonizedExpression = null;

	/**
	 * A string version of the canonized expression.
	 */
	private String canonizedString = null;

	/**
	 * The fresh part of the path condition together with the bounds.
	 */
	private Expression expressionBounded = null;

	/**
	 * The full expression with the fresh and non-fresh parts of the path
	 * condition.
	 */
	private Expression fullExpressionBounded = null;

	/**
	 * The sliced expression, including the variable bounds. If there is no
	 * active slicer, the full expression is used.
	 */
	private Expression slicedExpressionBounded = null;

	/**
	 * The canonized expression, including the variable bounds. If there is no
	 * active canonizer, the sliced expression is used.
	 */
	private Expression canonizedExpressionBounded = null;

	/**
	 * A string version of the canonized expression, including the variable
	 * bounds.
	 */
	private String canonizedStringBounded = null;

	/**
	 * A flag that stores whether or not the (variable-bounded) path condition is satisfiable.
	 */
	private Boolean satisfiableBounded = null;

	/**
	 * A flag that stores whether or not the path condition is satisfiable.
	 */
	private Boolean satisfiable = null;

	/**
	 * A stored counter for the number of satisfying path conditions.
	 */
	private Apint solutionCount = null;

	/**
	 * A map from variables that appear in the instance (in other words, path
	 * condition) to integer, real, and other kinds of values.
	 */
	private HashMap<Variable, Object> map = null;

	/**
	 * A map from variables to renamed variables.
	 */
	private HashMap<Variable, Variable> renaming = null;

	/**
	 * The set of all the variables that appear in the full expression.
	 */
	private Set<Variable> allVariables = null;

	/**
	 * The set of all the variables that appear in the sliced expression.
	 */
	private Set<Variable> slicedVariables = null;

	/**
	 * A constructor for instances. Initially all fields (except for those
	 * supplied as parameters) are set to <code>null</code> and they are only
	 * calculated when they are needed.
	 * 
	 * @param solver
	 * @param parent
	 * @param expression
	 */
	public Instance(Solver solver, Instance parent, Expression expression) {
		this.solver = solver;
		solver.instanceCount++;
		this.parent = parent;
		this.expression = expression;
		BoundsCollector collector = new BoundsCollector();
		expressionBounded = collector.addBounds(expression);
		fullExpression = null;
		slicedExpression = null;
		canonizedExpression = null;
		canonizedString = null;
		fullExpressionBounded = null;
		slicedExpressionBounded = null;
		canonizedExpressionBounded = null;
		canonizedStringBounded = null;
		satisfiableBounded = null;
		satisfiable = null;
		solutionCount = null;
		map = null;
		renaming = null;
		allVariables = null;
		slicedVariables = null;
	}

	/**
	 * Returns the parent instance.
	 * 
	 * @return the parent instance
	 */
	public Instance getParent() {
		return parent;
	}

	/**
	 * Returns the expression (= the fresh part of the path condition).
	 * 
	 * @return the expression
	 */
	public Expression getExpression() {
		return expression;
	}

	/**
	 * Returns the expression (= the fresh part of the path condition) with the
	 * variable bounds added.
	 * 
	 * @return the bounded expression
	 */
	public Expression getExpressionBounded() {
		return expressionBounded;
	}

	/**
	 * Returns the full expression that consists of the full expression of the
	 * parent conjoined with the expression of the current instance. If this
	 * conjunction is not yet available, it is calculated. It does not include
	 * the bounds on the variables.
	 * 
	 * @return the full expression for the path condition
	 */
	public Expression getFullExpression() {
		if (fullExpression == null) {
			Instance p = getParent();
			if (p == null) {
				fullExpression = getExpression();
			} else {
				fullExpression = new Operation(Operation.Operator.AND,
						getExpression(), p.getFullExpression());
			}
		}
		return fullExpression;
	}

	/**
	 * Returns the full expression that consists of the full expression of the
	 * parent conjoined with the expression of the current instance. If this
	 * conjunction is not yet available, it is calculated. It includes the
	 * bounds on the variables.
	 * 
	 * @return the full expression for the path condition including variable
	 *         bounds
	 */
	public Expression getFullExpressionBounded() {
		if (fullExpressionBounded == null) {
			Instance p = getParent();
			if (p == null) {
				fullExpressionBounded = getExpressionBounded();
			} else {
				fullExpressionBounded = new Operation(Operation.Operator.AND,
						getExpressionBounded(), p.getFullExpressionBounded());
			}
		}
		return fullExpressionBounded;
	}

	/**
	 * Returns the sliced expression for the path condition. If this is not yet
	 * available, it is calculated immediately. This calculation is based on the
	 * fresh part of the path condition. It does not include the bounds on the
	 * variables.
	 * 
	 * @return the sliced expression
	 */
	public Expression getSlicedExpression() {
		if (slicedExpression == null) {
			Slicer s = solver.getSlicer();
			if (s != null) {
				Instance p = getParent();
				if (p == null) {
					slicedExpression = getFullExpression();
				} else {
					slicedExpression = s.slice(getExpression(),
							p.getFullExpression());
				}
			} else {
				slicedExpression = getFullExpression();
			}
		}
		return slicedExpression;
	}

	/**
	 * Returns the sliced expression for the path condition. If this is not yet
	 * available, it is calculated immediately. This calculation is based on the
	 * fresh part of the path condition. It includes the bounds on the
	 * variables.
	 * 
	 * @return the sliced expression including variable bounds
	 */
	public Expression getSlicedExpressionBounded() {
		if (slicedExpressionBounded == null) {
			Slicer s = solver.getSlicer();
			if (s != null) {
				Instance p = getParent();
				if (p == null) {
					slicedExpressionBounded = getFullExpressionBounded();
				} else {
					slicedExpressionBounded = s.slice(getExpressionBounded(),
							p.getFullExpressionBounded());
				}
			} else {
				slicedExpressionBounded = getFullExpressionBounded();
			}
		}
		return slicedExpressionBounded;
	}

	/**
	 * Returns the canonized expression. If this is not yet available, it is
	 * calculated immediately. This calculation is based on the sliced
	 * expression. It does not include the bounds on the variables.
	 * 
	 * @return the canonized expression
	 */
	public Expression getCanonizedExpression() {
		if (canonizedExpression == null) {
			Canonizer c = solver.getCanonizer();
			if (c != null) {
				renaming = new HashMap<Variable, Variable>();
				canonizedExpression = c.canonize(getSlicedExpression(),
						renaming);
			} else {
				canonizedExpression = getSlicedExpression();
			}
		}
		return canonizedExpression;
	}

	/**
	 * Returns the canonized expression. If this is not yet available, it is
	 * calculated immediately. This calculation is based on the sliced
	 * expression. It includes the bounds on the variables.
	 * 
	 * @return the canonized expression including variable bounds
	 */
	public Expression getCanonizedExpressionBounded() {
		if (canonizedExpressionBounded == null) {
			Canonizer c = solver.getCanonizer();
			if (c != null) {
				renaming = new HashMap<Variable, Variable>();
				canonizedExpressionBounded = c.canonize(
						getSlicedExpressionBounded(), renaming);
			} else {
				canonizedExpressionBounded = getSlicedExpressionBounded();
			}
		}
		return canonizedExpressionBounded;
	}

	/**
	 * Returns a string version of the canonized expression. If not yet
	 * available, it is calculated immediately. It does not include the bounds
	 * on the variables.
	 * 
	 * @return a string version of the canonized expression
	 */
	public String getCanonizedString() {
		if (canonizedString == null) {
			Canonizer c = solver.getCanonizer();
			if (c != null) {
				canonizedString = c.canonizeToString(getCanonizedExpression());
			} else {
				canonizedString = getSlicedExpression().toString();
			}
		}
		return canonizedString;
	}

	/**
	 * Returns a string version of the canonized expression. If not yet
	 * available, it is calculated immediately. It includes the bounds on the
	 * variables.
	 * 
	 * @return a string version of the canonized expression including variable
	 *         bounds
	 */
	public String getCanonizedStringBounded() {
		if (canonizedStringBounded == null) {
			Canonizer c = solver.getCanonizer();
			if (c != null) {
				canonizedStringBounded = c
						.canonizeToString(getCanonizedExpressionBounded());
			} else {
				canonizedStringBounded = getSlicedExpressionBounded()
						.toString();
			}
		}
		return canonizedStringBounded;
	}

	public Set<Variable> getAllVariables() {
		if (allVariables == null) {
			Instance p = getParent();
			allVariables = new VariableCollector().getVariables(getExpression());
			if (p != null) {
				allVariables.addAll(p.getAllVariables());
			}
		}
		return allVariables;
	}

	/**
	 * Returns a set of all the variables that occur in the sliced expression.
	 * 
	 * @return a set of variables of the sliced expression
	 */
	public Set<Variable> getSlicedVariables() {
		if (slicedVariables == null) {
			slicedVariables = new VariableCollector().getVariables(getSlicedExpression());
		}
		return slicedVariables;
	}

	/**
	 * Returns a boolean to indicate whether or not the full instance expression
	 * is satisfiable, or, if there is no decision procedure available,
	 * <code>null</code>.
	 * 
	 * @return <code>null</code> if not decision procedure is available,
	 *         <code>true</code> if the full instance expression is satisfiable,
	 *         <code>false</code> otherwise
	 */
	
	static int counter = 0;
	
	public Boolean isSatisfiable() {
//		System.out.println("Instance.isSatisfiable()");
//		System.out.println("   Full expression :"+getFullExpression());
//		System.out.println("   Canonized expression :"+getCanonizedExpression());		
		if (satisfiable == null) {
			Store s = solver.getStore();
			String k = null;
			if (s != null) {
				k = "S:" + getCanonizedExpression();
				//System.out.println(counter + " Constraint = " + k);counter++;
				satisfiable = s.getBoolean(k);
				if (satisfiable != null) {
					solver.isSatStoreHitCount++;
					//System.out.println("   Reused result\n");
				}
				else {
					//System.out.println("   Solver called\n");
				}
			}
			if (satisfiable == null) {
				satisfiable = solver.getDecisionProcedure().isSatisfiable(
						getSlicedExpression());
				if (s != null) {
					s.put(k, satisfiable);
					solver.isSatStoreMissCount++;
				}
			}
		}
		return satisfiable;
	}

	/**
	 * Returns a boolean to indicate whether or not the full instance expression
	 * is satisfiable, or, if there is no decision procedure available,
	 * <code>null</code>.
	 * 
	 * @return <code>null</code> if not decision procedure is available,
	 *         <code>true</code> if the full instance expression is satisfiable,
	 *         <code>false</code> otherwise
	 */
	public Boolean isSatisfiableBounded() {
		if (satisfiableBounded == null) {
			Store s = solver.getStore();
			String k = null;
			if (s != null) {
				k = "S:" + getCanonizedExpressionBounded();
				satisfiableBounded = s.getBoolean(k);
				if (satisfiableBounded != null) {
					solver.isSatStoreHitCount++;
				}
			}
			if (satisfiableBounded == null) {
				satisfiableBounded = solver.getDecisionProcedure().isSatisfiable(
						getCanonizedExpressionBounded());
				if (s != null) {
					s.put(k, satisfiableBounded);
					solver.isSatStoreMissCount++;
				}
			}
		}
		return satisfiable;
	}

	/**
	 * Sets the satisfiability result for this instance. If it is already known
	 * or found in the store, the two results are asserted to be equal.
	 * Otherwise, the instance's satisfiability is set to the new value and it
	 * recorded in the store, if possible.
	 * 
	 * @param satisfiable
	 *            the new satisfiability result
	 */
	public void setSatisfiable(boolean satisfiable) {
		if (this.satisfiable == null) {
			Store s = solver.getStore();
			String k = null;
			if (s != null) {
				k = "S:" + getCanonizedExpression();
				this.satisfiable = s.getBoolean(k);
				if (this.satisfiable != null) {
					solver.isSatStoreHitCount++;
				}
			}
			if (this.satisfiable == null) {
				this.satisfiable = satisfiable;
				if (s != null) {
					s.put(k, this.satisfiable);
					solver.isSatStoreMissCount++;
				}
			}
		}
		assert this.satisfiable == satisfiable;
	}

	/**
	 * Sets the satisfiability result for this instance. If it is already known
	 * or found in the store, the two results are asserted to be equal.
	 * Otherwise, the instance's satisfiability is set to the new value and it
	 * recorded in the store, if possible.
	 * 
	 * @param satisfiable
	 *            the new satisfiability result
	 */
	public void setSatisfiableBounded(boolean satisfiable) {
		if (this.satisfiable == null) {
			Store s = solver.getStore();
			String k = null;
			if (s != null) {
				k = "S:" + getCanonizedExpressionBounded();
				this.satisfiable = s.getBoolean(k);
				if (this.satisfiable != null) {
					solver.isSatStoreHitCount++;
				}
			}
			if (this.satisfiable == null) {
				this.satisfiable = satisfiable;
				if (s != null) {
					s.put(k, this.satisfiable);
					solver.isSatStoreMissCount++;
				}
			}
		}
		assert this.satisfiable == satisfiable;
	}

	/**
	 * Returns the number of value assignments that satisfy the full instance
	 * expression, or, if there is no model counter available, <code>null</code>
	 * is returned.
	 * 
	 * @return <code>null</code> if no model counter is available, or the number
	 *         of value assignments that satisfy the full instance expression
	 */
	public Apint countSolutions() {
		if (solutionCount == null) {
			Store s = solver.getStore();
			String k = null;
			if (s != null) {
				k = "C:" + getCanonizedExpressionBounded();
				String st = s.getString(k);
				if (st != null) {
					solutionCount = new Apint(st);
				}
				if (solutionCount != null) {
					solver.countStoreHitCount++;
				}
			}
			if (solutionCount == null) {
				solutionCount = solver.getModelCounter().countSolutions(
						getCanonizedExpressionBounded());
				if (s != null) {
					s.put(k, solutionCount);
					solver.countStoreMissCount++;
				}
			}
		}
		return solutionCount;
	}

	/**
	 * Sets the solution count result for this instance. If it is already known
	 * or found in the store, the two results are asserted to be equal.
	 * Otherwise, the instance's solution count is set to the new value and it
	 * recorded in the store, if possible.
	 * 
	 * @param solutionCount
	 *            the new solution count
	 */
	public void setSolutionCount(Apint solutionCount) {
		if (this.solutionCount == null) {
			Store s = solver.getStore();
			String k = null;
			if (s != null) {
				k = "C:" + getCanonizedExpressionBounded();
				String st = s.getString(k);
				if (st != null) {
					this.solutionCount = new Apint(st);
				}
				if (this.solutionCount != null) {
					solver.countStoreHitCount++;
				}
			}
			if (this.solutionCount == null) {
				this.solutionCount = solutionCount;
				if (s != null) {
					s.put(k, this.solutionCount);
					solver.countStoreMissCount++;
				}
			}
		}
		assert this.solutionCount == solutionCount;
	}

	/**
	 * Returns the new name (or to be more precise, the new version) of a
	 * particular variable. If no map for renaming is available, the variable is
	 * returned as is.
	 * 
	 * @param variable
	 *            the variable to search for
	 * @return the new, renamed variable if available, or the old variable
	 *         itself
	 */
	private Variable getRenaming(Variable variable) {
		if (renaming == null) {
			return variable;
		} else {
			return renaming.get(variable);
		}
	}

	/**
	 * Returns the integer value that corresponds to a variable in a value
	 * assignment that satisfies the full instance expression. If either there
	 * is no constraint solver available, or the value of the variable does not
	 * matter, the functions returns 0.
	 * 
	 * @param variable
	 *            the integer variable
	 * @return the value corresponding to the variable or 0
	 */
	@SuppressWarnings("unchecked")
	public Integer getIntValue(IntVariable variable) {
		if (map == null) {
			Store s = solver.getStore();
			String k = null;
			if (s != null) {
				k = "M:" + getCanonizedExpressionBounded();
				map = (HashMap<Variable, Object>) s.get(k);
			}
			if (map == null) {
				map = solver.getConstraintSolver().getVariableMap(
						getCanonizedExpressionBounded());
				if (s != null) {
					s.put(k, map);
				}
			}
		}
		return (Integer) map.get(getRenaming(variable));
	}

	// added by bihuan
	public boolean findNextVariableMap() {
		HashMap<Variable, Object> next = solver.getConstraintSolver().getNextVariableMap();
		if (next == null) {
			return false;
		} else {
			map = next;
			return true;
		}
	}
	
	/**
	 * Returns the integer value that corresponds to a variable in a value
	 * assignment that satisfies the full instance expression. This is a
	 * convenience routine that accepts any variable and assumes that it is an
	 * integer variable.
	 * 
	 * @param variable
	 *            the variable (assumed to be a integer variable)
	 * @return the value of the variable or 0 if it has not been assigned a
	 *         value
	 */
	public Integer getIntValue(Variable variable) {
		return getIntValue((IntVariable) variable);
	}

	/**
	 * Returns the real value that corresponds to a variable in a value
	 * assignment that satisfies the full instance expression. If either there
	 * is no constraint solver available, or the value of the variable does not
	 * matter, the functions returns 0.
	 * 
	 * @param variable
	 *            the real variable
	 * @return the value corresponding to the variable or 0
	 */
	@SuppressWarnings("unchecked")
	public Double getRealValue(RealVariable variable) {
		if (map == null) {
			Store store = solver.getStore();
			String key = null;
			if (store != null) {
				key = "M:" + getCanonizedExpression();
				map = (HashMap<Variable, Object>) store.get(key);
			}
			if (map == null) {
				map = solver.getConstraintSolver().getVariableMap(
						getCanonizedExpression());
				if (store != null) {
					store.put(key, map);
				}
			}
		}
		return (Double) map.get(getRenaming(variable));
	}

	/**
	 * Returns the real value that corresponds to a variable in a value
	 * assignment that satisfies the full instance expression. This is a
	 * convenience routine that accepts any variable and assumes that it is a
	 * real variable.
	 * 
	 * @param variable
	 *            the variable (assumed to be a real variable)
	 * @return the value of the variable or 0 if it has not been assigned a
	 *         value
	 */
	public Double getRealValue(Variable variable) {
		return getRealValue((RealVariable) variable);
	}

	private static class BoundsCollector extends Visitor {

		private Expression bounds;

		public Expression addBounds(Expression expression) {
			bounds = null;
			expression.accept(this);
			if (bounds == null) {
				return expression;
			} else {
				return new Operation(Operation.Operator.AND, expression, bounds);
			}
		}

		@Override
		public void postVisit(IntVariable variable) {
			Expression boundConstraint = null;
			Integer bound = variable.getLowerBound();
			if (bound != null) {
				boundConstraint = new Operation(Operation.Operator.GE,
						variable, new IntConstant(bound));
				if (bounds == null) {
					bounds = boundConstraint;
				} else {
					bounds = new Operation(Operation.Operator.AND, bounds,
							boundConstraint);
				}
			}
			bound = variable.getUpperBound();
			if (bound != null) {
				boundConstraint = new Operation(Operation.Operator.LE,
						variable, new IntConstant(bound));
				if (bounds == null) {
					bounds = boundConstraint;
				} else {
					bounds = new Operation(Operation.Operator.AND, bounds,
							boundConstraint);
				}
			}
		}

	}

	private static class VariableCollector extends Visitor {

		private Set<Variable> variableSet;

		public Set<Variable> getVariables(Expression expression) {
			variableSet = new HashSet<Variable>();
			expression.accept(this);
			return variableSet;
		}

		@Override
		public void postVisit(Variable variable) {
			variableSet.add(variable);
		}

	}

}
