package za.ac.sun.cs.solver;

import java.util.ArrayList;
import java.util.List;

import za.ac.sun.cs.solver.canonization.Canonizer;
import za.ac.sun.cs.solver.csp.ConstraintSolver;
import za.ac.sun.cs.solver.csp.DecisionProcedure;
import za.ac.sun.cs.solver.csp.ModelCounter;
import za.ac.sun.cs.solver.slicing.Slicer;
import za.ac.sun.cs.solver.store.Store;

/**
 * A central manager for the solving of path conditions.
 * 
 * @author Jaco Geldenhuys <jaco@cs.sun.ac.za>
 */
public class Solver {

	/**
	 * The slicer that is used to minimize path conditions.
	 */
	private Slicer slicer;

	/**
	 * A canonizer that converts a path condition to a canonized form.
	 */
	private Canonizer canonizer;

	/**
	 * A decision procedure that determines whether or not the path condition is
	 * satisfiable.
	 */
	private DecisionProcedure dp;

	/**
	 * The constraint solver that finds value for variables that will satisfy a
	 * path condition.
	 */
	private ConstraintSolver cs;

	/**
	 * The model counter that counts the number of satisfying variable
	 * assignments that satisfy a path condition.
	 */
	private ModelCounter mc;

	/**
	 * The data store where results are cached.
	 */
	private Store store;

	/**
	 * The number of instances that refer to this solver.
	 */
	public int instanceCount = 0;

	/**
	 * The number of times a satisfiability result is found in the store.
	 */
	public int isSatStoreHitCount = 0;

	/**
	 * The number of times a satisfiability result is not found in the store.
	 */
	public int isSatStoreMissCount = 0;

	/**
	 * The number of times a solution count result is found in the store.
	 */
	public int countStoreHitCount = 0;

	/**
	 * The number of times a solution count result is not found in the store.
	 */
	public int countStoreMissCount = 0;

	/**
	 * Constructor that initializes the fields of the solver.
	 */
	public Solver() {
		slicer = null;
		canonizer = null;
		dp = null;
		cs = null;
		mc = null;
		store = null;
	}

	/**
	 * Returns a boolean to indicate whether or not slicing is active.
	 * 
	 * @return <code>true</code> if slicing is active; <code>false</code>
	 *         otherwise
	 */
	public boolean isSlicing() {
		return slicer != null;
	}

	/**
	 * Returns the slicer.
	 * 
	 * @return the slicer
	 */
	public Slicer getSlicer() {
		return slicer;
	}

	/**
	 * Sets a new value for the slicer.
	 * 
	 * @param slicer
	 *            the new slicer
	 */
	public void setSlicer(Slicer slicer) {
		this.slicer = slicer;
	}

	/**
	 * Returns a boolean to indicate whether or not canonization is active.
	 * 
	 * @return <code>true</code> if canonization is active; <code>false</code>
	 *         otherwise
	 */
	public boolean isCanonizing() {
		return canonizer != null;
	}

	/**
	 * Returns the canonizer.
	 * 
	 * @return the canonizer
	 */
	public Canonizer getCanonizer() {
		return canonizer;
	}

	/**
	 * Sets a new value for the canonizer.
	 * 
	 * @param canonizer
	 *            the new canonizer
	 */
	public void setCanonizer(Canonizer canonizer) {
		this.canonizer = canonizer;
	}

	/**
	 * Returns the decision procedure.
	 * 
	 * @return the decision procedure
	 */
	public DecisionProcedure getDecisionProcedure() {
		return dp;
	}

	/**
	 * Sets a new value for the decision procedure.
	 * 
	 * @param dp
	 *            the new decision procedure
	 */
	public void setDecisionProcedure(DecisionProcedure dp) {
		this.dp = dp;
	}

	/**
	 * Returns the constraint solver.
	 * 
	 * @return the constraint solver
	 */
	public ConstraintSolver getConstraintSolver() {
		return cs;
	}

	/**
	 * Sets a new value for the constraint solver.
	 * 
	 * @param cs
	 *            the new constraint solver
	 */
	public void setConstraintSolver(ConstraintSolver cs) {
		this.cs = cs;
	}

	/**
	 * Returns the model counter.
	 * 
	 * @return the model counter
	 */
	public ModelCounter getModelCounter() {
		return mc;
	}

	/**
	 * Sets a new value for the model counter.
	 * 
	 * @param mc
	 *            the new model counter
	 */
	public void setModelCounter(ModelCounter mc) {
		this.mc = mc;
	}

	/**
	 * Returns a boolean to indicate whether or not storage is active.
	 * 
	 * @return <code>true</code> if storage is active; <code>false</code>
	 *         otherwise
	 */
	public boolean isStoring() {
		return store != null;
	}

	/**
	 * Returns the store.
	 * 
	 * @return the store
	 */
	public Store getStore() {
		return store;
	}

	/**
	 * Sets a new value for the store.
	 * 
	 * @param store
	 *            the new store
	 */
	public void setStore(Store store) {
		this.store = store;
	}

	/**
	 * Shuts down the solver.
	 * 
	 * At the moment, this shuts down the store only. In the future, however, it
	 * may be necessary to shut down other components of the solver as well.
	 */
	public void shutdown() {
		if (store != null) {
			store.shutdown();
		}
	}

	/**
	 * Returns a report on the performance of the solver as a list of strings.
	 * 
	 * @return a list of strings that constitute a performance report
	 */
	public List<String> getSolverReport() {
		List<String> report = new ArrayList<String>();
		report.add(">>>--- solver report begin --->>>");
		report.add("solver instanceCount : " + instanceCount);
		if (store != null) {
			if (dp != null) {
				report.add("solver isSatStoreHitCount : " + isSatStoreHitCount);
				report.add("solver isSatStoreMissCount : "
						+ isSatStoreMissCount);
			}
			if (mc != null) {
				report.add("solver countStoreHitCount : " + countStoreHitCount);
				report.add("solver countStoreMissCount : "
						+ countStoreMissCount);
			}
		}
		if (slicer != null) {
			report.addAll(slicer.getSlicerReport());
		}
		if (canonizer != null) {
			report.addAll(canonizer.getCanonizerReport());
		}
		if (dp != null) {
			report.addAll(dp.getDecisionProcedureReport());
		}
		if (cs != null) {
			report.addAll(cs.getConstraintSolverReport());
		}
		if (mc != null) {
			report.addAll(mc.getModelCounterReport());
		}
		if (store != null) {
			report.addAll(store.getStoreReport());
		}
		report.add(">>>--- solver report end --->>>");
		return report;
	}

}
