package za.ac.sun.cs.solver.canonization;

import java.util.List;
import java.util.Map;

import za.ac.sun.cs.solver.expr.Expression;
import za.ac.sun.cs.solver.expr.Variable;

/**
 * An interface that provides for the canonization of expressions. Canonization
 * means that expressions are converted to a standard form.
 * 
 * For example, <code>x&lt;y</code> and <code>y&gt;x</code> are clearly
 * equivalent. But so is <code>x-y&lt;0</code>. But from a satisfiability point
 * of view, it is also equivalent to <code>z&lt;q</code> and even
 * <code>y&lt;x</code>.
 * 
 * When it comes to constraint solving, however,
 * <code>x&lt;y<&code> and <code>y&lt;x</code> are equivalent as long as long as
 * the system keeps track of variable renaming.
 * 
 * @author Jaco Geldenhuys <jaco@cs.sun.ac.za>
 */
public interface Canonizer {

	/**
	 * Returns the canonized version of the given expression while populating
	 * the map with variable renamings.
	 * 
	 * @param expression
	 *            the expression to canonize
	 * @param map
	 *            a map populated with variable renamings
	 * @return a canonized version of the expression
	 */
	public Expression canonize(Expression expression,
			Map<Variable, Variable> map);

	/**
	 * Returns a string version of the canonization of the given expression.
	 * 
	 * @param expression
	 *            the expression to canonize
	 * @return a canonized string version of the expression
	 */
	public String canonizeToString(Expression expression);

	/**
	 * Returns a list of "report" lines about the performance of the canonizer.
	 * 
	 * @return lines of a performance report
	 */
	public List<String> getCanonizerReport();

}
