package za.ac.sun.cs.solver.canonization;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;

import za.ac.sun.cs.solver.expr.Constant;
import za.ac.sun.cs.solver.expr.Expression;
import za.ac.sun.cs.solver.expr.IntConstant;
import za.ac.sun.cs.solver.expr.IntVariable;
import za.ac.sun.cs.solver.expr.Operation;
import za.ac.sun.cs.solver.expr.Variable;
import za.ac.sun.cs.solver.expr.Visitor;

public class DefaultCanonizer implements Canonizer {

	/**
	 * Number of times the slicer has been invoked.
	 */
	private int invocations = 0;

	@Override
	public Expression canonize(Expression expression,
			Map<Variable, Variable> map) {
		invocations++;
		OrderingVisitor orderingVisitor = new OrderingVisitor();
		expression.accept(orderingVisitor);
		expression = orderingVisitor.getExpression();
		CanonizationVisitor canonizationVisitor = new CanonizationVisitor();
		expression.accept(canonizationVisitor);
		Expression canonized = canonizationVisitor.getExpression();
		if (canonized != null) {
			canonized = new Renamer(map, canonizationVisitor.getVariableSet())
					.rename(canonized);
		}
		return canonized;
	}

	@Override
	public String canonizeToString(Expression expression) {
		Printer printer = new Printer();
		expression.accept(printer);
		return printer.getString();
	}

	private static class OrderingVisitor extends Visitor {

		private Stack<Expression> stack;

		public OrderingVisitor() {
			stack = new Stack<Expression>();
		}

		public Expression getExpression() {
			return stack.pop();
		}

		@Override
		public void postVisit(IntConstant constant) {
			stack.push(constant);
		}

		@Override
		public void postVisit(IntVariable variable) {
			stack.push(variable);
		}

		@Override
		public void postVisit(Operation operation) {
			Operation.Operator op = operation.getOperator();
			Operation.Operator nop = null;
			switch (op) {
			case EQ:
				nop = Operation.Operator.EQ;
				break;
			case NE:
				nop = Operation.Operator.NE;
				break;
			case LT:
				nop = Operation.Operator.GT;
				break;
			case LE:
				nop = Operation.Operator.GE;
				break;
			case GT:
				nop = Operation.Operator.LT;
				break;
			case GE:
				nop = Operation.Operator.LE;
				break;
			}
			if (nop != null) {
				Expression r = stack.pop();
				Expression l = stack.pop();
				if ((r instanceof IntVariable) && (l instanceof IntVariable) && (((IntVariable) r).getName().compareTo(((IntVariable) l).getName()) < 0)) {
					stack.push(new Operation(nop, r, l));
				} else if ((r instanceof IntVariable) && (l instanceof IntConstant)) {
					stack.push(new Operation(nop, r, l));
				} else {
					stack.push(operation);
				}
			} else if (op.getArity() == 2) {
				Expression r = stack.pop();
				Expression l = stack.pop();
				stack.push(new Operation(op, l, r));
			} else {
				for (int i = op.getArity(); i > 0; i--) {
					stack.pop();
				}
				stack.push(operation);
			}
		}

	}

	private static class CanonizationVisitor extends Visitor {

		private static final IntConstant ZERO = new IntConstant(0);

		private static final IntConstant ONE = new IntConstant(1);

		private static final Expression FALSE = new Operation(
				Operation.Operator.EQ, ZERO, ONE);

		private static final Expression TRUE = new Operation(
				Operation.Operator.EQ, ZERO, ZERO);

		private Stack<Expression> stack;

		private SortedSet<Expression> conjuncts;

		private SortedSet<IntVariable> variableSet;

		private Map<IntVariable, Integer> lowerBounds;

		private Map<IntVariable, Integer> upperBounds;

		private IntVariable boundVariable;

		private Integer bound;

		private int boundCoeff;

		private boolean unsatisfiable;

		private boolean linearInteger;

		public CanonizationVisitor() {
			stack = new Stack<Expression>();
			conjuncts = new TreeSet<Expression>();
			variableSet = new TreeSet<IntVariable>();
			unsatisfiable = false;
			linearInteger = true;
		}

		public SortedSet<IntVariable> getVariableSet() {
			return variableSet;
		}

		public Expression getExpression() {
			if (!linearInteger) {
				return null;
			} else if (unsatisfiable) {
				return FALSE;
			} else {
				if (!stack.isEmpty()) {
					Expression x = stack.pop();
					if (!x.equals(TRUE)) {
						conjuncts.add(x);
					}
				}
				SortedSet<Expression> newConjuncts = processBounds();
				new TreeSet<Expression>();
				Expression c = null;
				for (Expression e : newConjuncts) {
					Operation o = (Operation) e;
					if (o.getOperator() == Operation.Operator.GT) {
						e = new Operation(Operation.Operator.LT, scale(-1,
								o.getOperand(0)), o.getOperand(1));
					} else if (o.getOperator() == Operation.Operator.GE) {
						e = new Operation(Operation.Operator.LE, scale(-1,
								o.getOperand(0)), o.getOperand(1));
					}
					o = (Operation) e;
					if (o.getOperator() == Operation.Operator.GT) {
						e = new Operation(Operation.Operator.GE, merge(
								o.getOperand(0), new IntConstant(-1)),
								o.getOperand(1));
					} else if (o.getOperator() == Operation.Operator.LT) {
						e = new Operation(Operation.Operator.LE, merge(
								o.getOperand(0), new IntConstant(1)),
								o.getOperand(1));
					}
					if (c == null) {
						c = e;
					} else {
						c = new Operation(Operation.Operator.AND, c, e);
					}
				}
				return c;
			}
		}

		private SortedSet<Expression> processBounds() {
			return conjuncts;
			// SortedSet<Expression> newConjuncts = new TreeSet<Expression>();
			// lowerBounds = new HashMap<IntVariable, Integer>();
			// upperBounds = new HashMap<IntVariable, Integer>();
			// for (Expression e : conjuncts) {
			// extractBound(e);
			// }
			// for (Expression e : conjuncts) {
			// Expression f = boundExpression(e);
			// if (f != null) {
			// newConjuncts.add(f);
			// }
			// }
			// return newConjuncts;
		}

		@SuppressWarnings("unused")
		private void extractBound(Expression e) {
			if (e instanceof Operation) {
				Operation o = (Operation) e;
				Expression lhs = o.getOperand(0);
				Operation.Operator op = o.getOperator();
				if (isBound(lhs)) {
					switch (op) {
					case EQ:
						lowerBounds.put(boundVariable, bound * boundCoeff);
						upperBounds.put(boundVariable, bound * boundCoeff);
						break;
					case LT:
						if (boundCoeff == 1) {
							upperBounds.put(boundVariable, bound * -1 - 1);
						} else {
							lowerBounds.put(boundVariable, bound + 1);
						}
						break;
					case LE:
						if (boundCoeff == 1) {
							upperBounds.put(boundVariable, bound * -1);
						} else {
							lowerBounds.put(boundVariable, bound);
						}
						break;
					case GT:
						if (boundCoeff == 1) {
							lowerBounds.put(boundVariable, bound * -1 + 1);
						} else {
							upperBounds.put(boundVariable, bound - 1);
						}
						break;
					case GE:
						if (boundCoeff == 1) {
							lowerBounds.put(boundVariable, bound * -1);
						} else {
							upperBounds.put(boundVariable, bound);
						}
						break;
					}
				}
			}
		}

		private boolean isBound(Expression lhs) {
			if (!(lhs instanceof Operation)) {
				return false;
			}
			Operation o = (Operation) lhs;
			if (o.getOperator() == Operation.Operator.MUL) {
				if (!(o.getOperand(0) instanceof IntConstant)) {
					return false;
				}
				if (!(o.getOperand(1) instanceof IntVariable)) {
					return false;
				}
				boundVariable = (IntVariable) o.getOperand(1);
				bound = 0;
				if ((((IntConstant) o.getOperand(0)).getValue() == 1)
						|| (((IntConstant) o.getOperand(0)).getValue() == -1)) {
					boundCoeff = ((IntConstant) o.getOperand(0)).getValue();
					return true;
				} else {
					return false;
				}
			} else if (o.getOperator() == Operation.Operator.ADD) {
				if (!(o.getOperand(1) instanceof IntConstant)) {
					return false;
				}
				bound = ((IntConstant) o.getOperand(1)).getValue();
				if (!(o.getOperand(0) instanceof Operation)) {
					return false;
				}
				Operation p = (Operation) o.getOperand(0);
				if (!(p.getOperand(0) instanceof IntConstant)) {
					return false;
				}
				if (!(p.getOperand(1) instanceof IntVariable)) {
					return false;
				}
				boundVariable = (IntVariable) p.getOperand(1);
				if ((((IntConstant) p.getOperand(0)).getValue() == 1)
						|| (((IntConstant) p.getOperand(0)).getValue() == -1)) {
					boundCoeff = ((IntConstant) p.getOperand(0)).getValue();
					return true;
				} else {
					return false;
				}
			} else {
				return false;
			}
		}

		@Override
		public void postVisit(Constant constant) {
			if (linearInteger && !unsatisfiable) {
				if (constant instanceof IntConstant) {
					stack.push(constant);
				} else {
					stack.clear();
					linearInteger = false;
				}
			}
		}

		@Override
		public void postVisit(Variable variable) {
			if (linearInteger && !unsatisfiable) {
				if (variable instanceof IntVariable) {
					variableSet.add((IntVariable) variable);
					stack.push(new Operation(Operation.Operator.MUL, ONE,
							variable));
				} else {
					stack.clear();
					linearInteger = false;
				}
			}
		}

		@Override
		public void postVisit(Operation operation) {
			if (!linearInteger || unsatisfiable) {
				return;
			}
			Operation.Operator op = operation.getOperator();
			switch (op) {
			case AND:
				if (!stack.isEmpty()) {
					Expression x = stack.pop();
					if (!x.equals(TRUE)) {
						conjuncts.add(x);
					}
				}
				if (!stack.isEmpty()) {
					Expression x = stack.pop();
					if (!x.equals(TRUE)) {
						conjuncts.add(x);
					}
				}
				break;
			case EQ:
			case NE:
			case LT:
			case LE:
			case GT:
			case GE:
				Expression e = merge(scale(-1, stack.pop()), stack.pop());
				if (e instanceof IntConstant) {
					int v = ((IntConstant) e).getValue();
					boolean b = true;
					if (op == Operation.Operator.EQ) {
						b = v == 0;
					} else if (op == Operation.Operator.NE) {
						b = v != 0;
					} else if (op == Operation.Operator.LT) {
						b = v < 0;
					} else if (op == Operation.Operator.LE) {
						b = v <= 0;
					} else if (op == Operation.Operator.GT) {
						b = v > 0;
					} else if (op == Operation.Operator.GE) {
						b = v >= 0;
					}
					if (b) {
						stack.push(TRUE);
					} else {
						unsatisfiable = true;
					}
				} else {
					stack.push(new Operation(op, e, ZERO));
				}
				break;
			case ADD:
				stack.push(merge(stack.pop(), stack.pop()));
				break;
			case SUB:
				stack.push(merge(scale(-1, stack.pop()), stack.pop()));
				break;
			case MUL:
				Expression r = stack.pop();
				Expression l = stack.pop();
				if ((l instanceof IntConstant) && (r instanceof IntConstant)) {
					int li = ((IntConstant) l).getValue();
					int ri = ((IntConstant) r).getValue();
					stack.push(new IntConstant(li * ri));
				} else if (l instanceof IntConstant) {
					int li = ((IntConstant) l).getValue();
					stack.push(scale(li, r));
				} else if (r instanceof IntConstant) {
					int ri = ((IntConstant) r).getValue();
					stack.push(scale(ri, l));
				} else {
					stack.clear();
					linearInteger = false;
				}
			}
		}

		private Expression merge(Expression left, Expression right) {
			Operation l = null;
			Operation r = null;
			int s = 0;
			if (left instanceof IntConstant) {
				s = ((IntConstant) left).getValue();
			} else {
				if (hasRightConstant(left)) {
					s = getRightConstant(left);
					l = getLeftOperation(left);
				} else {
					l = (Operation) left;
				}
			}
			if (right instanceof IntConstant) {
				s += ((IntConstant) right).getValue();
			} else {
				if (hasRightConstant(right)) {
					s += getRightConstant(right);
					r = getLeftOperation(right);
				} else {
					r = (Operation) right;
				}
			}
			SortedMap<Variable, Integer> coefficients = new TreeMap<Variable, Integer>();
			IntConstant c;
			Variable v;
			Integer k;

			// Collect the coefficients of l
			if (l != null) {
				while (l.getOperator() == Operation.Operator.ADD) {
					Operation o = (Operation) l.getOperand(1);
					assert (o.getOperator() == Operation.Operator.MUL);
					c = (IntConstant) o.getOperand(0);
					v = (IntVariable) o.getOperand(1);
					coefficients.put(v, c.getValue());
					l = (Operation) l.getOperand(0);
				}
				assert (l.getOperator() == Operation.Operator.MUL);
				c = (IntConstant) l.getOperand(0);
				v = (IntVariable) l.getOperand(1);
				coefficients.put(v, c.getValue());
			}

			// Collect the coefficients of r
			if (r != null) {
				while (r.getOperator() == Operation.Operator.ADD) {
					Operation o = (Operation) r.getOperand(1);
					assert (o.getOperator() == Operation.Operator.MUL);
					c = (IntConstant) o.getOperand(0);
					v = (IntVariable) o.getOperand(1);
					k = coefficients.get(v);
					if (k == null) {
						coefficients.put(v, c.getValue());
					} else {
						coefficients.put(v, c.getValue() + k);
					}
					r = (Operation) r.getOperand(0);
				}
				assert (r.getOperator() == Operation.Operator.MUL);
				c = (IntConstant) r.getOperand(0);
				v = (IntVariable) r.getOperand(1);
				k = coefficients.get(v);
				if (k == null) {
					coefficients.put(v, c.getValue());
				} else {
					coefficients.put(v, c.getValue() + k);
				}
			}

			Expression lr = null;
			for (Map.Entry<Variable, Integer> e : coefficients.entrySet()) {
				int coef = e.getValue();
				if (coef != 0) {
					Operation term = new Operation(Operation.Operator.MUL,
							new IntConstant(coef), e.getKey());
					if (lr == null) {
						lr = term;
					} else {
						lr = new Operation(Operation.Operator.ADD, lr, term);
					}
				}
			}
			if ((lr == null) || (lr instanceof IntConstant)) {
				return new IntConstant(s);
			} else if (s == 0) {
				return lr;
			} else {
				return new Operation(Operation.Operator.ADD, lr,
						new IntConstant(s));
			}
		}

		private boolean hasRightConstant(Expression expression) {
			return isAddition(expression)
					&& (getRightExpression(expression) instanceof IntConstant);
		}

		private int getRightConstant(Expression expression) {
			return ((IntConstant) getRightExpression(expression)).getValue();
		}

		private Expression getLeftExpression(Expression expression) {
			return ((Operation) expression).getOperand(0);
		}

		private Expression getRightExpression(Expression expression) {
			return ((Operation) expression).getOperand(1);
		}

		private Operation getLeftOperation(Expression expression) {
			return (Operation) getLeftExpression(expression);
		}

		// private Operation getRightOperation(Expression expression) {
		// return (Operation) getRightExpression(expression);
		// }

		private boolean isAddition(Expression expression) {
			return ((Operation) expression).getOperator() == Operation.Operator.ADD;
		}

		private Expression scale(int factor, Expression expression) {
			if (factor == 0) {
				return ZERO;
			}
			if (expression instanceof IntConstant) {
				return new IntConstant(factor
						* ((IntConstant) expression).getValue());
			} else if (expression instanceof IntVariable) {
				return expression;
			} else {
				assert (expression instanceof Operation);
				Operation o = (Operation) expression;
				Operation.Operator p = o.getOperator();
				Expression l = scale(factor, o.getOperand(0));
				Expression r = scale(factor, o.getOperand(1));
				return new Operation(p, l, r);
			}
		}

	}

	private static class Renamer extends Visitor {

		private Map<Variable, Variable> map;

		private Stack<Expression> stack;

		public Renamer(Map<Variable, Variable> map,
				SortedSet<IntVariable> variableSet) {
			// for (IntVariable v : variableSet) {
			// map.put(v, new IntVariable("v" + map.size(), v.getLowerBound(),
			// v.getUpperBound()));
			// }
			this.map = map;
			stack = new Stack<Expression>();
		}

		public Expression rename(Expression expression) {
			expression.accept(this);
			return stack.pop();
		}

		@Override
		public void postVisit(IntVariable variable) {
			Variable v = map.get(variable);
			if (v == null) {
				v = new IntVariable("v" + map.size(), variable.getLowerBound(),
						variable.getUpperBound());
				map.put(variable, v);
			}
			stack.push(v);
		}

		@Override
		public void postVisit(IntConstant constant) {
			stack.push(constant);
		}

		@Override
		public void postVisit(Operation operation) {
			int arity = operation.getOperator().getArity();
			Expression operands[] = new Expression[arity];
			for (int i = arity; i > 0; i--) {
				operands[i - 1] = stack.pop();
			}
			stack.push(new Operation(operation.getOperator(), operands));
		}

	}

	private static class Printer extends Visitor {

		private StringBuilder stringBuilder;

		public Printer() {
			stringBuilder = new StringBuilder();
		}

		public String getString() {
			return stringBuilder.toString();
		}

		@Override
		public void postVisit(Constant constant) {
			stringBuilder.append(constant.toString());
			stringBuilder.append(';');
		}

		@Override
		public void postVisit(Variable variable) {
			stringBuilder.append(variable.getName());
			stringBuilder.append(';');
		}

		@Override
		public void postVisit(Operation operation) {
			stringBuilder.append(operation.getOperator().toString());
		}

	}

	@Override
	public List<String> getCanonizerReport() {
		List<String> report = new ArrayList<String>();
		report.add("canonizer invocations : " + invocations);
		return report;
	}
}
