package za.ac.sun.cs.solver.csp;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import choco.cp.model.CPModel;
import choco.cp.solver.CPSolver;
import choco.cp.solver.search.integer.valselector.RandomIntValSelector;
import choco.cp.solver.search.integer.varselector.RandomIntVarSelector;
import choco.kernel.model.constraints.Constraint;
import choco.kernel.model.variables.integer.IntegerExpressionVariable;
import choco.kernel.model.variables.integer.IntegerVariable;
import za.ac.sun.cs.solver.expr.Expression;
import za.ac.sun.cs.solver.expr.IntConstant;
import za.ac.sun.cs.solver.expr.IntVariable;
import za.ac.sun.cs.solver.expr.Operation;
import za.ac.sun.cs.solver.expr.Variable;
import za.ac.sun.cs.solver.expr.Visitor;

public class Choco implements DecisionProcedure, ConstraintSolver {

	private Expression expression = null;

	private Boolean satisfiable = null;

	private HashMap<Variable, Object> model = null;

	private Map<Variable, IntegerVariable> variableMap = null;

	private CPModel chocoModel = null;

	private CPSolver chocoSolver = null;

	private int isSatInvocationCount = 0;

	private long dpTimeConsumption = 0;

	private long csTimeConsumption = 0;

	public Choco() {
		expression = null;
		satisfiable = null;
		model = null;
	}

	private void solve(Expression expression) {
		chocoModel = new CPModel();
		this.expression = expression;
		variableMap = new HashMap<Variable, IntegerVariable>();
		new Translator(chocoModel, variableMap).translate(expression);
		chocoSolver = new CPSolver();
		chocoSolver.read(chocoModel);
		//chocoSolver.setRandomSelectors(); // added by bihuan
		chocoSolver.setValIntSelector(new RandomIntValSelector()); // added by bihuan
		chocoSolver.setVarIntSelector(new RandomIntVarSelector(chocoSolver)); // added by bihuan
		satisfiable = chocoSolver.solve();
		model = null;
	}

	@Override
	public Boolean isSatisfiable(Expression expression) {
		isSatInvocationCount++;
		if (this.expression == null || !this.expression.equals(expression)) {
			long startTime = System.currentTimeMillis();
			solve(expression);
			dpTimeConsumption += System.currentTimeMillis() - startTime;
		}
		return satisfiable;
	}

	@Override
	public HashMap<Variable, Object> getVariableMap(Expression expression) {
		if (this.expression == null || !this.expression.equals(expression)) {
			long startTime = System.currentTimeMillis();
			solve(expression);
			csTimeConsumption += System.currentTimeMillis() - startTime;
		}
		if (model == null) {
			model = new HashMap<Variable, Object>(); // added by bihuan
			long startTime = System.currentTimeMillis();
			for (Map.Entry<Variable, IntegerVariable> e : variableMap.entrySet()) {
				model.put(e.getKey(), chocoSolver.getVar(e.getValue()).getVal());
			}
			csTimeConsumption += System.currentTimeMillis() - startTime;
		}
		return model;
	}

	// added by bihuan
	public HashMap<Variable, Object> getNextVariableMap() {
		if (chocoSolver.nextSolution()) {
			long startTime = System.currentTimeMillis();
			model = new HashMap<Variable, Object>();
			for (Map.Entry<Variable, IntegerVariable> e : variableMap.entrySet()) {
				model.put(e.getKey(), chocoSolver.getVar(e.getValue()).getVal());
			}
			csTimeConsumption += System.currentTimeMillis() - startTime;
			return model;
		} else {
			return null;
		}
	}
	
	@Override
	public Integer getIntValue(Variable variable) {
		if (this.expression == null || !this.expression.equals(expression)) {
			long startTime = System.currentTimeMillis();
			solve(expression);
			csTimeConsumption += System.currentTimeMillis() - startTime;
		}
		IntegerVariable v = variableMap.get(variable);
		return (v == null) ? null : chocoSolver.getVar(v).getVal();
	}

	@Override
	public Double getRealValue(Variable variable) {
		return null;
	}

	private static class Translator extends Visitor {

		private CPModel chocoModel = null;

		private Stack<Object> stack = null;

		private List<Constraint> constraints = null;

		private Map<Variable, IntegerVariable> variableMap = null;

		public Translator(CPModel chocoModel, Map<Variable, IntegerVariable> variableMap) {
			this.chocoModel = chocoModel;
			stack = new Stack<Object>();
			constraints = new LinkedList<Constraint>();
			this.variableMap = variableMap;
		}

		public void translate(Expression expression) {
			expression.accept(this);
			for (Constraint c : constraints) {
				chocoModel.addConstraint(c);
			}
			if (!stack.isEmpty()) {
				chocoModel.addConstraint((Constraint) stack.pop());
			}
			assert stack.isEmpty();
		}

		@Override
		public void postVisit(IntConstant constant) {
			stack.push(constant.getValue());
		}

		@Override
		public void postVisit(IntVariable variable) {
			IntegerVariable v = variableMap.get(variable);
			if (v == null) {
				Integer lower = variable.getLowerBound();
				Integer upper = variable.getUpperBound();
				v = choco.Choco.makeIntVar(variable.getName(), lower, upper);
				chocoModel.addVariable(v);
				variableMap.put(variable, v);
			}
			stack.push(v);
		}

		@Override
		public void postVisit(Operation operation) {
			Object l = null;
			Object r = null;
			int arity = operation.getOperator().getArity();
			if (arity == 2) {
				if (!stack.isEmpty()) {
					r = stack.pop();
				}
				if (!stack.isEmpty()) {
					l = stack.pop();
				}
			} else if (arity == 1) {
				if (!stack.isEmpty()) {
					l = stack.pop();
				}
			}
			switch (operation.getOperator()) {
			case EQ:
				if (l instanceof Integer) {
					stack.push(choco.Choco.eq((Integer) l, (IntegerExpressionVariable) r));
				} else if (r instanceof Integer) {
					stack.push(choco.Choco.eq((IntegerExpressionVariable) l, (Integer) r));
				} else {
					stack.push(choco.Choco.eq((IntegerExpressionVariable) l, (IntegerExpressionVariable) r));
				}
				break;
			case NE:
				if (l instanceof Integer) {
					stack.push(choco.Choco.neq((Integer) l, (IntegerExpressionVariable) r));
				} else if (r instanceof Integer) {
					stack.push(choco.Choco.neq((IntegerExpressionVariable) l, (Integer) r));
				} else {
					stack.push(choco.Choco.neq((IntegerExpressionVariable) l, (IntegerExpressionVariable) r));
				}
				break;
			case LT:
				if (l instanceof Integer) {
					stack.push(choco.Choco.lt((Integer) l, (IntegerExpressionVariable) r));
				} else if (r instanceof Integer) {
					stack.push(choco.Choco.lt((IntegerExpressionVariable) l, (Integer) r));
				} else {
					stack.push(choco.Choco.lt((IntegerExpressionVariable) l, (IntegerExpressionVariable) r));
				}
				break;
			case LE:
				if (l instanceof Integer) {
					stack.push(choco.Choco.leq((Integer) l, (IntegerExpressionVariable) r));
				} else if (r instanceof Integer) {
					stack.push(choco.Choco.leq((IntegerExpressionVariable) l, (Integer) r));
				} else {
					stack.push(choco.Choco.leq((IntegerExpressionVariable) l, (IntegerExpressionVariable) r));
				}
				break;
			case GT:
				if (l instanceof Integer) {
					stack.push(choco.Choco.gt((Integer) l, (IntegerExpressionVariable) r));
				} else if (r instanceof Integer) {
					stack.push(choco.Choco.gt((IntegerExpressionVariable) l, (Integer) r));
				} else {
					stack.push(choco.Choco.gt((IntegerExpressionVariable) l, (IntegerExpressionVariable) r));
				}
				break;
			case GE:
				if (l instanceof Integer) {
					stack.push(choco.Choco.geq((Integer) l, (IntegerExpressionVariable) r));
				} else if (r instanceof Integer) {
					stack.push(choco.Choco.geq((IntegerExpressionVariable) l, (Integer) r));
				} else {
					stack.push(choco.Choco.geq((IntegerExpressionVariable) l, (IntegerExpressionVariable) r));
				}
				break;
			case AND:
				if (l != null) {
					constraints.add((Constraint) l);
				}
				if (r != null) {
					constraints.add((Constraint) r);
				}
				break;
			case ADD:
				if (l instanceof Integer) {
					stack.push(choco.Choco.plus((Integer) l, (IntegerExpressionVariable) r));
				} else if (r instanceof Integer) {
					stack.push(choco.Choco.plus((IntegerExpressionVariable) l, (Integer) r));
				} else {
					stack.push(choco.Choco.plus((IntegerExpressionVariable) l, (IntegerExpressionVariable) r));
				}
				break;
			case SUB:
				if (l instanceof Integer) {
					stack.push(choco.Choco.minus((Integer) l, (IntegerExpressionVariable) r));
				} else if (r instanceof Integer) {
					stack.push(choco.Choco.minus((IntegerExpressionVariable) l, (Integer) r));
				} else {
					stack.push(choco.Choco.minus((IntegerExpressionVariable) l, (IntegerExpressionVariable) r));
				}
				break;
			case MUL:
				stack.push(choco.Choco.mult((Integer) l, (IntegerExpressionVariable) r));
				break;
			case IMPLIES:
			case OR:
			case DIV:
			case MOD:
			case BIT_AND:
			case BIT_OR:
			case BIT_XOR:
			case SHIFTL:
			case SHIFTR:
			case SHIFTUR:
			case SIN:
			case COS:
			case TAN:
			case ASIN:
			case ACOS:
			case ATAN:
			case ATAN2:
			case ROUND:
			case LOG:
			case EXP:
			case POWER:
			case SQRT:
			default:
				throw new RuntimeException("## Error: operation not supported");
			}
		}
	}

	@Override
	public List<String> getConstraintSolverReport() {
		List<String> report = new ArrayList<String>();
		report.add("choco isSatisfiableInvocations : " + isSatInvocationCount);
		report.add("choco dpTimeConsumption : " + dpTimeConsumption);
		return report;
	}

	@Override
	public List<String> getDecisionProcedureReport() {
		List<String> report = new ArrayList<String>();
		report.add("choco csTimeConsumption : " + csTimeConsumption);
		return report;
	}

}
