package za.ac.sun.cs.solver.csp;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;


// need Z3 JNI bindings can we reuse the one from Scala Z3?
import cvc3.Expr;
import cvc3.FlagsMut;
import cvc3.Rational;
import cvc3.SatResult;
import cvc3.Type;
import cvc3.ValidityChecker;
import za.ac.sun.cs.solver.expr.Expression;
import za.ac.sun.cs.solver.expr.IntConstant;
import za.ac.sun.cs.solver.expr.IntVariable;
import za.ac.sun.cs.solver.expr.Operation;
import za.ac.sun.cs.solver.expr.RealConstant;
import za.ac.sun.cs.solver.expr.RealVariable;
import za.ac.sun.cs.solver.expr.Variable;
import za.ac.sun.cs.solver.expr.Visitor;

public class Z3 implements DecisionProcedure, ConstraintSolver {

	private static final int BASE = 10; // used in creating real variables

	private Expression expression = null;

	private Boolean satisfiable = null;

	private HashMap<Variable, Object> model = null;

	private int isSatInvocationCount = 0;

	private long dpTimeConsumption = 0;

	private long csTimeConsumption = 0;

	public Z3() {
		expression = null;
		satisfiable = null;
		model = null;
	}

	private void solve(Expression expression) {
		FlagsMut flags = ValidityChecker.createFlags(null);
		flags.setFlag("dagify-exprs", false);
		ValidityChecker validityChecker = ValidityChecker.create(flags);
		this.expression = expression;
		Translator translator = new Translator(validityChecker);
		expression.accept(translator);
		long startTime = System.currentTimeMillis();
		validityChecker.push();
		SatResult result = validityChecker.checkUnsat(translator
				.getTranslation());
		dpTimeConsumption += System.currentTimeMillis() - startTime;
		if (result == SatResult.SATISFIABLE) {
			startTime = System.currentTimeMillis();
			@SuppressWarnings("rawtypes")
			HashMap m = null;
			try {
				m = validityChecker.getConcreteModel();
			} catch (Exception x) {
				// Sometimes Z3 produces an exception here. Why?
				System.out.println("############ Z3 Exception ###########");
				System.out.println("############ PC: " + expression);
				x.printStackTrace();
			}
			validityChecker.pop();
			csTimeConsumption += System.currentTimeMillis() - startTime;
			if (m != null) {
				model = new HashMap<Variable, Object>();
				for (Map.Entry<Variable, Expr> e : translator.getMap()
						.entrySet()) {
					if (e.getKey() instanceof IntVariable) {
						Expr x = (Expr) m.get((Expr) e.getValue());
						model.put(e.getKey(), x.getRational().getInteger());
					} else {
						Expr x = (Expr) m.get((Expr) e.getValue());
						Rational r = x.getRational();
						double num = r.getNumerator().getInteger();
						double den = r.getDenominator().getInteger();
						model.put(e.getKey(), num / den);
					}
				}
			}
			satisfiable = true;
		} else {
			validityChecker.pop();
			model = null;
			satisfiable = false;
		}
		if (validityChecker != null) {
			validityChecker.delete();
		}
		if (flags != null) {
			flags.delete();
		}
		validityChecker = null;
		flags = null;
	}

	@Override
	public Boolean isSatisfiable(Expression expression) {
		isSatInvocationCount++;
		if (satisfiable == null || !this.expression.equals(expression)) {
			solve(expression);
		}
		return satisfiable;
	}

	@Override
	public HashMap<Variable, Object> getVariableMap(Expression expression) {
		if (satisfiable == null || !this.expression.equals(expression)
				|| (model == null)) {
			solve(expression);
		}
		return model;
	}

	@Override
	public Integer getIntValue(Variable variable) {
		if (satisfiable == null || (model == null)) {
			solve(expression);
		}
		return (Integer) model.get(variable);
	}

	@Override
	public Double getRealValue(Variable variable) {
		if (satisfiable == null || (model == null)) {
			solve(expression);
		}
		return (Double) model.get(variable);
	}

	private static class Translator extends Visitor {

		private ValidityChecker validityChecker = null;

		private Stack<Expr> stack = null;

		private Map<Variable, Expr> v2e = null;

		public Translator(ValidityChecker validityChecker) {
			this.validityChecker = validityChecker;
			stack = new Stack<Expr>();
			v2e = new HashMap<Variable, Expr>();
		}

		public Expr getTranslation() {
			return stack.pop();
		}

		public Map<Variable, Expr> getMap() {
			return v2e;
		}

		@Override
		public void postVisit(IntConstant constant) {
			stack.push(validityChecker.ratExpr(constant.getValue()));
		}

		@Override
		public void postVisit(RealConstant constant) {
			stack.push(validityChecker.ratExpr(
					Double.toString(constant.getValue()), BASE));
		}

		@Override
		public void postVisit(IntVariable variable) {
			Expr v = v2e.get(variable);
			if (v == null) {
				Integer lower = variable.getLowerBound();
				Integer upper = variable.getUpperBound();
				Type type = validityChecker.subrangeType(
						validityChecker.ratExpr(lower),
						validityChecker.ratExpr(upper));
				v = validityChecker.varExpr(variable.getName(), type);
				v2e.put(variable, v);
			}
			stack.push(v);
		}

		@Override
		public void postVisit(RealVariable variable) {
			Expr v = v2e.get(variable);
			if (v == null) {
				int lower = (int) (double) variable.getLowerBound();
				int upper = (int) (double) variable.getUpperBound();
				Type type = validityChecker.subrangeType(
						validityChecker.ratExpr(lower),
						validityChecker.ratExpr(upper));
				v = validityChecker.varExpr(variable.getName(), type);
				v2e.put(variable, v);
			}
			stack.push(v);
		}

		@Override
		public void postVisit(Operation operation) {
			Expr l = null;
			Expr r = null;
			int arity = operation.getOperator().getArity();
			if (arity == 2) {
				if (!stack.isEmpty()) {
					r = stack.pop();
				}
				if (!stack.isEmpty()) {
					l = stack.pop();
				}
			} else if (arity == 1) {
				if (!stack.isEmpty()) {
					l = stack.pop();
				}
			}
			switch (operation.getOperator()) {
			case EQ:
				stack.push(validityChecker.eqExpr(l, r));
				break;
			case NE:
				stack.push(validityChecker.notExpr(validityChecker.eqExpr(l, r)));
				break;
			case LT:
				stack.push(validityChecker.ltExpr(l, r));
				break;
			case LE:
				stack.push(validityChecker.leExpr(l, r));
				break;
			case GT:
				stack.push(validityChecker.gtExpr(l, r));
				break;
			case GE:
				stack.push(validityChecker.geExpr(l, r));
				break;
			case AND:
				stack.push(validityChecker.andExpr(l, r));
				break;
			case OR:
				stack.push(validityChecker.orExpr(l, r));
				break;
			case IMPLIES:
				stack.push(validityChecker.impliesExpr(l, r));
				break;
			case ADD:
				stack.push(validityChecker.plusExpr(l, r));
				break;
			case SUB:
				stack.push(validityChecker.minusExpr(l, r));
				break;
			case MUL:
				stack.push(validityChecker.multExpr(l, r));
				break;
			case DIV:
				stack.push(validityChecker.divideExpr(l, r));
				break;
			case MOD:
			case BIT_AND:
			case BIT_OR:
			case BIT_XOR:
			case SHIFTL:
			case SHIFTR:
			case SHIFTUR:
			case SIN:
			case COS:
			case TAN:
			case ASIN:
			case ACOS:
			case ATAN:
			case ATAN2:
			case ROUND:
			case LOG:
			case EXP:
			case POWER:
			case SQRT:
			default:
				throw new RuntimeException("## Error: operation not supported");
			}
		}
	}

	@Override
	public List<String> getConstraintSolverReport() {
		List<String> report = new ArrayList<String>();
		report.add("cvc3 isSatisfiableInvocations : " + isSatInvocationCount);
		report.add("cvc3 dpTimeConsumption : " + dpTimeConsumption);
		return report;
	}

	@Override
	public List<String> getDecisionProcedureReport() {
		List<String> report = new ArrayList<String>();
		report.add("cvc3 csTimeConsumption : " + csTimeConsumption);
		return report;
	}

	@Override
	public HashMap<Variable, Object> getNextVariableMap() {
		// TODO Auto-generated method stub
		return null;
	}

}
