package za.ac.sun.cs.solver.store;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import redis.clients.jedis.Jedis;

/**
 * An implementation of a {@link Store} based on redis (
 * <code>http://www.redis.io</code>).
 * 
 * @author Jaco Geldenhuys <jaco@cs.sun.ac.za>
 */
public class RedisLevelStore extends Store {

	private static final class StringPair {
		
		private String first = null;
		
		private String second = null;

		public StringPair(String first, String second) {
			this.first = first;
			this.second = second;
		}

		public String getFirst() {
			return first;
		}

		public String getSecond() {
			return second;
		}
	}

	private final static String TERMINATE = "TERMINATE";

	private final class Writer implements Runnable {
		
		private final BlockingQueue<StringPair> queue;

		public Writer(BlockingQueue<StringPair> queue) {
			this.queue = queue;
		}

		public void run() {
			try {
				while (true) {
					StringPair p = queue.take();
					if (p.getFirst().equals(TERMINATE)) {
						break;
					}
					for (int i = 1; i < levelCount; i++) {
						writedb[i].set(p.getFirst(), p.getSecond());
					}
				}
			} catch (InterruptedException e) {
				System.out
						.println("#### Exception in queue: " + e.getMessage());
				e.printStackTrace();
				throw new RuntimeException();
			}
		}
	}

	/**
	 * The time (in seconds) of inactivity until the connection to the redis
	 * store timeout.
	 */
	private static final int TIMEOUT = 2000;

	/**
	 * Connection to the redis stores.
	 */
	private Jedis[] db = null;

	/**
	 * Second set of connections for the writing thread.
	 */
	private Jedis[] writedb = null;

	/**
	 * The shared queue where store write requests are recorded.
	 */
	private BlockingQueue<StringPair> writeQueue = null;

	/**
	 * The thread that performs the writing.
	 */
	private Thread writeThread = null;

	/**
	 * The number of redis stores available.
	 */
	private int levelCount = 0;

	/**
	 * Number of times <code>get(...)</code> was called.
	 */
	private int[] retrievalCount = null;

	/**
	 * Number of times <code>put(...)</code> was called.
	 */
	private int[] insertionCount = null;

	/**
	 * Constructor to create a default connection to a redis store running on
	 * the local computer.
	 */
	public RedisLevelStore() {
		levelCount = 1;
		db = new Jedis[levelCount];
		db[0] = new Jedis("localhost", 6379, TIMEOUT);
		retrievalCount = new int[levelCount];
		insertionCount = new int[levelCount];
		for (int i = 0; i < levelCount; i++) {
			retrievalCount[i] = 0;
			insertionCount[i] = 0;
		}
		writeQueue = new LinkedBlockingQueue<StringPair>();
		writeThread = new Thread(new Writer(writeQueue));
		writeThread.start();
	}

	/**
	 * Constructor to create a connection to a redis store given the host and
	 * the port.
	 * 
	 * @param host
	 *            the host on which the redis store is running
	 * @param port
	 *            the port on which the redis store is listening
	 */
	public RedisLevelStore(String[] host, int[] port) {
		levelCount = host.length;
		if (port.length > levelCount) {
			levelCount = port.length;
		}
		if (levelCount == 0) {
			levelCount = 1;
			db = new Jedis[levelCount];
			db[0] = new Jedis("localhost", 6379, TIMEOUT);
		} else {
			db = new Jedis[levelCount];
			writedb = new Jedis[levelCount];
			for (int i = 0; i < levelCount; i++) {
				String h = (i < host.length) ? host[i] : "localhost";
				int p = (i < port.length) ? port[i] : 6379;
				db[i] = new Jedis(h, p, TIMEOUT);
				writedb[i] = new Jedis(h, p, TIMEOUT);
			}
		}
		retrievalCount = new int[levelCount];
		insertionCount = new int[levelCount];
		for (int i = 0; i < levelCount; i++) {
			retrievalCount[i] = 0;
			insertionCount[i] = 0;
		}
		writeQueue = new LinkedBlockingQueue<StringPair>();
		writeThread = new Thread(new Writer(writeQueue));
		writeThread.start();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see za.ac.sun.cs.solver.store.Store#shutdown()
	 */
	@Override
	public void shutdown() {
		try {
			writeQueue.put(new StringPair(TERMINATE, null));
		} catch (InterruptedException e) {
			System.out.println("#### Exception in queue: " + e.getMessage());
			e.printStackTrace();
			throw new RuntimeException();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see za.ac.sun.cs.solver.store.Store#getString(java.lang.String)
	 */
	@Override
	public String getString(String key) {
		for (int i = 0; i < levelCount; i++) {
			retrievalCount[i]++;
			String value = db[i].get(key);
			if (value != null) {
				// Propagate the found value to all other stores, even if some of the stores may already contain this information.
				put(key, value);
				return value;
			}
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see za.ac.sun.cs.solver.store.Store#put(java.lang.String,
	 * java.io.Serializable)
	 */
	@Override
	public void put(String key, Serializable value) {
		insertionCount[0]++;
		db[0].set(key, value.toString());
		try {
			writeQueue.put(new StringPair(key, value.toString()));
		} catch (InterruptedException e) {
			System.out.println("#### Exception in queue: " + e.getMessage());
			e.printStackTrace();
			throw new RuntimeException();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see za.ac.sun.cs.solver.store.Store#put(java.lang.String,
	 * java.io.Serializable)
	 */
	@Override
	public void put(String key, String value) {
		insertionCount[0]++;
		db[0].set(key, value);
		try {
			writeQueue.put(new StringPair(key, value));
		} catch (InterruptedException e) {
			System.out.println("#### Exception in queue: " + e.getMessage());
			e.printStackTrace();
			throw new RuntimeException();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see za.ac.sun.cs.solver.store.Store#getStoreReport()
	 */
	@Override
	public List<String> getStoreReport() {
		List<String> report = new ArrayList<String>();
		for (int i = 0; i < levelCount; i++) {
			report.add("store retrievals[" + i + "] : " + retrievalCount[i]);
			report.add("store insertions[" + i + "] : " + insertionCount[i]);
		}
		return report;
	}

}
