package za.ac.sun.cs.solver.store;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import redis.clients.jedis.Jedis;

/**
 * An implementation of a {@link Store} based on redis (<code>http://www.redis.io</code>).
 * 
 * @author Jaco Geldenhuys <jaco@cs.sun.ac.za>
 */
public class RedisStore extends Store {

	/**
	 * The time (in seconds) of inactivity until the connection to the redis store timeout.
	 */
	private static final int TIMEOUT = 2000;
	
	/**
	 * Connection to the redis store.
	 */
	private Jedis db = null;

	/**
	 * Map where delayed insertions are stored.
	 */
	private Map<String, String> delayStore = null;

	/**
	 * Number of times <code>get(...)</code> was called.
	 */
	private int retrievalCount = 0;

	/**
	 * Number of times <code>put(...)</code> was called.
	 */
	private int insertionCount = 0;

	/**
	 * Number of times a retrieval was found in the delay store.
	 */
	private int delayRetrievalCount = 0;

	/**
	 * Constructor to create a default connection to a redis store running on the local computer.
	 */
	public RedisStore() {
		db = new Jedis("localhost", 6379, TIMEOUT);
	}

	/**
	 * Constructor to create a connection to a redis store given the host and the port.
	 * 
	 * @param host the host on which the redis store is running
	 * @param port the port on which the redis store is listening
	 */
	public RedisStore(String host, int port) {
		db = new Jedis(host, port, TIMEOUT);
	}

	/* (non-Javadoc)
	 * @see za.ac.sun.cs.solver.store.Store#shutdown()
	 */
	@Override
	public void shutdown() {
		if (delayStore != null) {
			for (Map.Entry<String, String> e : delayStore.entrySet()) {
				db.set(e.getKey(), e.getValue());
			}
		}
	}

	/**
	 * Sets whether a delayed store is used.
	 * 
	 * @param delay flag to indicate that writes should be delayed
	 */
	public void setDelayedWrites(boolean delay) {
		if (delay) {
			delayStore = new HashMap<String, String>();
		} else {
			delayStore = null;
		}
	}

	/* (non-Javadoc)
	 * @see za.ac.sun.cs.solver.store.Store#getString(java.lang.String)
	 */
	@Override
	public String getString(String key) {
		retrievalCount++;
		if (delayStore != null) {
			String s = delayStore.get(key);
			if (s != null) {
				delayRetrievalCount++;
				return s;
			}
		}
		//if (db.get(key) != null)
		//	System.out.println("@@@@@@@ " + key);
		return db.get(key);
	}

	/* (non-Javadoc)
	 * @see za.ac.sun.cs.solver.store.Store#put(java.lang.String, java.io.Serializable)
	 */
	@Override
	public void put(String key, Serializable value) {
		insertionCount++;
		if (delayStore != null) {
			delayStore.put(key, value.toString());
		} else {
			db.set(key, value.toString());
		}
	}

	/* (non-Javadoc)
	 * @see za.ac.sun.cs.solver.store.Store#put(java.lang.String, java.io.Serializable)
	 */
	@Override
	public void put(String key, String value) {
		insertionCount++;
		if (delayStore != null) {
			delayStore.put(key, value);
		} else {
			db.set(key, value);
		}
	}

	/* (non-Javadoc)
	 * @see za.ac.sun.cs.solver.store.Store#getStoreReport()
	 */
	@Override
	public List<String> getStoreReport() {
		List<String> report = new ArrayList<String>();
		if (delayStore != null) {
			report.add("store self-retrievals : " + delayRetrievalCount);
		}
		report.add("store retrievals : " + retrievalCount);
		report.add("store insertions : " + insertionCount);
		return report;
	}
}
